# From: https://raw.githubusercontent.com/google/gemma_pytorch/main/gemma/model.py

# Copyright 2024 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Inference-only Gemma model implementation."""

import re
import torch
from torch import nn
import torch.nn.functional as F
from typing import Any, List, Optional, Sequence, Tuple, Union

from . import config as gemma_config
from . import tokenizer


class Sampler(nn.Module):

  def __init__(self, vocab_size: int):
    super().__init__()
    self.vocab_size = vocab_size

  @torch.no_grad()
  def forward(
      self,
      embedding: torch.Tensor,
      hidden_states: torch.Tensor,
      output_positions: torch.Tensor,
      temperatures: torch.Tensor,
      top_ps: torch.Tensor,
      top_ks: torch.Tensor,
      embedding_bias: Optional[torch.Tensor] = None,
  ) -> torch.Tensor:
    # Select the last element for each sequence.
    # (batch_size, input_len, hidden_size) -> (batch_size, hidden_size)
    hidden_states = hidden_states.index_select(1,
                                               output_positions).squeeze(dim=1)
    logits = torch.matmul(hidden_states, embedding.t())
    if embedding_bias is not None:
      logits += embedding_bias

    if temperatures is None:
      return torch.argmax(logits, dim=-1).squeeze(dim=-1)

    # Apply temperature scaling.
    logits.div_(temperatures.unsqueeze(dim=1))

    # Calculate probabilities with softmax.
    probs = torch.softmax(logits, dim=-1, dtype=torch.float)
    probs_sort, probs_idx = torch.sort(probs, dim=-1, descending=True)

    # Apply top-p, top-k.
    probs_sum = torch.cumsum(probs_sort, dim=-1)
    top_ps_mask = (probs_sum - probs_sort) > top_ps.unsqueeze(dim=1)
    probs_sort = torch.where(top_ps_mask, 0, probs_sort)

    top_ks_mask = torch.arange(probs_idx.shape[-1], device=probs_idx.device)
    top_ks_mask = top_ks_mask.expand(probs_idx.shape[0], -1)
    top_ks_mask = top_ks_mask >= top_ks.unsqueeze(dim=1)
    probs_sort = torch.where(top_ks_mask, 0, probs_sort)

    # Re-normalization.
    probs_sort.div_(probs_sort.sum(dim=-1, keepdim=True))
    probs = torch.gather(
        probs_sort, dim=-1, index=torch.argsort(probs_idx, dim=-1))
    next_token_ids = torch.multinomial(
        probs, num_samples=1, replacement=True).squeeze(dim=-1)
    return next_token_ids


def precompute_freqs_cis(dim: int,
                         end: int,
                         theta: float = 10000.0) -> torch.Tensor:
  """Precomputes the frequency cis."""
  freqs = 1.0 / (theta**(torch.arange(0, dim, 2)[:(dim // 2)].float() / dim))
  t = torch.arange(end, device=freqs.device)
  freqs = torch.outer(t, freqs).float()
  freqs_cis = torch.polar(torch.ones_like(freqs), freqs)  # complex64
  return freqs_cis


def apply_rotary_emb(x: torch.Tensor, freqs_cis: torch.Tensor) -> torch.Tensor:
  """Applies the rotary embedding to the query and key tensors."""
  x_ = torch.view_as_complex(
      torch.stack(torch.chunk(x.transpose(1, 2).float(), 2, dim=-1), dim=-1))
  x_out = torch.view_as_real(x_ * freqs_cis).type_as(x)
  x_out = torch.cat(torch.chunk(x_out, 2, dim=-1), dim=-2)
  x_out = x_out.reshape(x_out.shape[0], x_out.shape[1], x_out.shape[2],
                        -1).transpose(1, 2)
  return x_out


class Linear(nn.Module):

  def __init__(self, in_features: int, out_features: int, quant: bool):
    super().__init__()
    if quant:
      self.weight = nn.Parameter(
          torch.empty((out_features, in_features), dtype=torch.int8),
          requires_grad=False,
      )
      self.weight_scaler = nn.Parameter(torch.Tensor(out_features))
    else:
      self.weight = nn.Parameter(
          torch.empty((out_features, in_features)),
          requires_grad=False,
      )
    self.quant = quant

  def forward(self, x):
    weight = self.weight
    if self.quant:
      weight = weight * self.weight_scaler.unsqueeze(-1)
    output = F.linear(x, weight)
    return output


class Embedding(nn.Module):

  def __init__(self, num_embeddings: int, embedding_dim: int, quant: bool):
    super().__init__()
    if quant:
      self.weight = nn.Parameter(
          torch.empty((num_embeddings, embedding_dim), dtype=torch.int8),
          requires_grad=False,
      )
      self.weight_scaler = nn.Parameter(torch.Tensor(num_embeddings))
    else:
      self.weight = nn.Parameter(
          torch.empty((num_embeddings, embedding_dim)),
          requires_grad=False,
      )
    self.quant = quant

  def forward(self, x):
    weight = self.weight
    if self.quant:
      weight = weight * self.weight_scaler.unsqueeze(-1)
    output = F.embedding(x, weight)
    return output


class RMSNorm(torch.nn.Module):

  def __init__(
      self,
      dim: int,
      eps: float = 1e-6,
      add_unit_offset: bool = True,
  ):
    super().__init__()
    self.eps = eps
    self.add_unit_offset = add_unit_offset
    self.weight = nn.Parameter(torch.zeros(dim))

  def _norm(self, x):
    return x * torch.rsqrt(x.pow(2).mean(-1, keepdim=True) + self.eps)

  def forward(self, x):
    x = self._norm(x.float()).type_as(x)
    if self.add_unit_offset:
      output = x * (1 + self.weight)
    else:
      output = x * self.weight
    return output


class GemmaMLP(nn.Module):

  def __init__(
      self,
      hidden_size: int,
      intermediate_size: int,
      quant: bool,
  ):
    super().__init__()
    self.gate_proj = Linear(hidden_size, intermediate_size, quant)
    self.up_proj = Linear(hidden_size, intermediate_size, quant)
    self.down_proj = Linear(intermediate_size, hidden_size, quant)

  def forward(self, x):
    gate = self.gate_proj(x)
    gate = F.gelu(gate)
    up = self.up_proj(x)
    fuse = gate * up
    outputs = self.down_proj(fuse)
    return outputs


class GemmaAttention(nn.Module):

  def __init__(
      self,
      hidden_size: int,
      num_heads: int,
      num_kv_heads: int,
      head_dim: int,
      quant: bool,
  ):
    super().__init__()

    self.num_heads = num_heads
    self.num_kv_heads = num_kv_heads

    assert self.num_heads % self.num_kv_heads == 0
    self.num_queries_per_kv = self.num_heads // self.num_kv_heads

    self.hidden_size = hidden_size
    self.head_dim = head_dim

    self.q_size = self.num_heads * self.head_dim
    self.kv_size = self.num_kv_heads * self.head_dim

    self.scaling = self.head_dim**-0.5

    self.qkv_proj = Linear(
        self.hidden_size,
        (self.num_heads + 2 * self.num_kv_heads) * self.head_dim,
        quant=quant)
    self.o_proj = Linear(
        self.num_heads * self.head_dim, self.hidden_size, quant=quant)

  def forward(
      self,
      hidden_states: torch.Tensor,
      freqs_cis: torch.Tensor,
      kv_write_indices: torch.Tensor,
      kv_cache: Tuple[torch.Tensor, torch.Tensor],
      mask: torch.Tensor,
  ) -> torch.Tensor:
    hidden_states_shape = hidden_states.shape
    assert len(hidden_states_shape) == 3

    batch_size, input_len, _ = hidden_states_shape

    qkv = self.qkv_proj(hidden_states)
    xq, xk, xv = qkv.split([self.q_size, self.kv_size, self.kv_size], dim=-1)

    xq = xq.view(batch_size, -1, self.num_heads, self.head_dim)
    xk = xk.view(batch_size, -1, self.num_kv_heads, self.head_dim)
    xv = xv.view(batch_size, -1, self.num_kv_heads, self.head_dim)

    # Positional embedding.
    xq = apply_rotary_emb(xq, freqs_cis=freqs_cis)
    xk = apply_rotary_emb(xk, freqs_cis=freqs_cis)

    # Write new kv cache.
    # [batch_size, input_len, n_local_kv_heads, head_dim]
    k_cache, v_cache = kv_cache
    k_cache.index_copy_(1, kv_write_indices, xk)
    v_cache.index_copy_(1, kv_write_indices, xv)

    key = k_cache
    value = v_cache
    if self.num_kv_heads != self.num_heads:
      # [batch_size, max_seq_len, n_local_heads, head_dim]
      key = torch.repeat_interleave(key, self.num_queries_per_kv, dim=2)
      value = torch.repeat_interleave(value, self.num_queries_per_kv, dim=2)

    # [batch_size, n_local_heads, input_len, head_dim]
    q = xq.transpose(1, 2)
    # [batch_size, n_local_heads, max_seq_len, head_dim]
    k = key.transpose(1, 2)
    v = value.transpose(1, 2)

    # [batch_size, n_local_heads, input_len, max_seq_len]
    scores = torch.matmul(q, k.transpose(2, 3)) * self.scaling
    scores = scores + mask
    scores = F.softmax(scores.float(), dim=-1).type_as(q)

    # [batch_size, n_local_heads, input_len, head_dim]
    output = torch.matmul(scores, v)

    # [batch_size, input_len, hidden_dim]
    output = (
        output.transpose(1, 2).contiguous().view(batch_size, input_len, -1))
    output = self.o_proj(output)
    return output


class GemmaDecoderLayer(nn.Module):

  def __init__(
      self,
      config: gemma_config.GemmaConfig,
  ):
    super().__init__()
    self.self_attn = GemmaAttention(
        hidden_size=config.hidden_size,
        num_heads=config.num_attention_heads,
        num_kv_heads=config.num_key_value_heads,
        head_dim=config.head_dim,
        quant=config.quant,
    )
    self.mlp = GemmaMLP(
        hidden_size=config.hidden_size,
        intermediate_size=config.intermediate_size,
        quant=config.quant,
    )
    self.input_layernorm = RMSNorm(config.hidden_size, eps=config.rms_norm_eps)
    self.post_attention_layernorm = RMSNorm(
        config.hidden_size, eps=config.rms_norm_eps)

  def forward(
      self,
      hidden_states: torch.Tensor,
      freqs_cis: torch.Tensor,
      kv_write_indices: torch.Tensor,
      kv_cache: Tuple[torch.Tensor, torch.Tensor],
      mask: torch.Tensor,
  ) -> torch.Tensor:
    # Self Attention
    residual = hidden_states
    hidden_states = self.input_layernorm(hidden_states)
    hidden_states = self.self_attn(
        hidden_states=hidden_states,
        freqs_cis=freqs_cis,
        kv_write_indices=kv_write_indices,
        kv_cache=kv_cache,
        mask=mask,
    )
    hidden_states = residual + hidden_states

    # MLP
    residual = hidden_states
    hidden_states = self.post_attention_layernorm(hidden_states)
    hidden_states = self.mlp(hidden_states)
    hidden_states = residual + hidden_states

    return hidden_states


class GemmaModel(nn.Module):

  def __init__(self, config: gemma_config.GemmaConfig):
    super().__init__()
    self.config = config
    self.vocab_size = config.vocab_size

    self.layers = nn.ModuleList()
    for _ in range(config.num_hidden_layers):
      self.layers.append(GemmaDecoderLayer(config))
    self.norm = RMSNorm(config.hidden_size, eps=config.rms_norm_eps)

  def forward(
      self,
      hidden_states: torch.Tensor,
      freqs_cis: torch.Tensor,
      kv_write_indices: torch.Tensor,
      kv_caches: List[Tuple[torch.Tensor, torch.Tensor]],
      mask: torch.Tensor,
  ) -> torch.Tensor:
    for i in range(len(self.layers)):
      layer = self.layers[i]
      hidden_states = layer(
          hidden_states=hidden_states,
          freqs_cis=freqs_cis,
          kv_write_indices=kv_write_indices,
          kv_cache=kv_caches[i],
          mask=mask,
      )
    hidden_states = self.norm(hidden_states)
    return hidden_states


class GemmaForCausalLM(nn.Module):

  def __init__(
      self,
      config: gemma_config.GemmaConfig,
  ):
    super().__init__()
    self.config = config
    assert config.hidden_size % config.num_attention_heads == 0

    max_seq_len = config.max_position_embeddings
    head_dim = config.head_dim
    vocab_size = config.vocab_size

    self.tokenizer = None  #tokenizer.Tokenizer(config.tokenizer)
    self.embedder = Embedding(vocab_size, config.hidden_size, config.quant)
    self.model = GemmaModel(config)
    self.sampler = Sampler(vocab_size)

    # Pre-compute rotary embedding table.
    rope_theta = getattr(config, 'rope_theta', 10000)
    freqs_cis = precompute_freqs_cis(
        head_dim, max_seq_len * 2, theta=rope_theta)
    self.register_buffer('freqs_cis', freqs_cis)

  @torch.no_grad()
  def forward(
      self,
      input_token_ids: torch.Tensor,
      input_positions: torch.Tensor,
      kv_write_indices: torch.Tensor,
      kv_caches: List[Tuple[torch.Tensor, torch.Tensor]],
      mask: torch.Tensor,
      output_positions: torch.Tensor,
      temperatures: torch.Tensor,
      top_ps: torch.Tensor,
      top_ks: torch.Tensor,
      **kwargs,
  ) -> torch.Tensor:
    freqs_cis = self.freqs_cis.index_select(0, input_positions)
    kv_write_indices = input_positions

    # [batch_size, input_len, hidden_size]
    hidden_states = self.embedder(input_token_ids)
    # Gemma normalizes the embedding by sqrt(hidden_size).
    hidden_states = hidden_states * (self.config.hidden_size**0.5)

    hidden_states = self.model(
        hidden_states=hidden_states,
        freqs_cis=freqs_cis,
        kv_write_indices=kv_write_indices,
        kv_caches=kv_caches,
        mask=mask,
    )
    embedder_weight = self.embedder.weight
    if self.config.quant:
      embedder_weight = (
          embedder_weight * self.embedder.weight_scaler.unsqueeze(-1))
    #next_tokens = self.sampler(
    return hidden_states
    # return next_tokens

  def generate(
      self,
      prompts: Union[str, Sequence[str]],
      device: Any,
      output_len: int = 100,
      temperature: float = 0.95,
      top_p: float = 1.0,
      top_k: int = 100,
  ) -> Union[str, Sequence[str]]:
    """Generates responses for given prompts using Gemma model."""
    # If a single prompt is provided, treat it as a batch of 1.
    is_str_prompt = isinstance(prompts, str)
    if is_str_prompt:
      prompts = [prompts]

    batch_size = len(prompts)
    prompt_tokens = [self.tokenizer.encode(prompt) for prompt in prompts]
    min_prompt_len = min(len(p) for p in prompt_tokens)
    max_prompt_len = max(len(p) for p in prompt_tokens)
    max_seq_len = max_prompt_len + output_len
    assert max_seq_len <= self.config.max_position_embeddings

    # build KV caches
    kv_caches = []
    for _ in range(self.config.num_hidden_layers):
      size = (batch_size, max_seq_len, self.config.num_key_value_heads,
              self.config.head_dim)
      dtype = self.config.get_dtype()
      k_cache = torch.zeros(size=size, dtype=dtype, device=device)
      v_cache = torch.zeros(size=size, dtype=dtype, device=device)
      kv_caches.append((k_cache, v_cache))

    # prepare inputs
    token_ids_tensor = torch.full((batch_size, max_seq_len),
                                  self.tokenizer.pad_id,
                                  dtype=torch.int64)
    input_token_ids_tensor = torch.full((batch_size, min_prompt_len),
                                        self.tokenizer.pad_id,
                                        dtype=torch.int64)
    for i, p in enumerate(prompt_tokens):
      token_ids_tensor[i, :len(p)] = torch.tensor(p)
      input_token_ids_tensor[i, :min_prompt_len] = torch.tensor(
          p[:min_prompt_len])
    token_ids_tensor = token_ids_tensor.to(device)
    input_token_ids_tensor = input_token_ids_tensor.to(device)
    prompt_mask_tensor = token_ids_tensor != self.tokenizer.pad_id
    input_positions_tensor = torch.arange(
        0, min_prompt_len, dtype=torch.int64).to(device)
    mask_tensor = torch.full((1, 1, max_seq_len, max_seq_len),
                             -2.3819763e38).to(torch.float)
    mask_tensor = torch.triu(mask_tensor, diagonal=1).to(device)
    curr_mask_tensor = mask_tensor.index_select(2, input_positions_tensor)
    output_positions_tensor = torch.LongTensor([min_prompt_len - 1]).to(device)
    temperatures_tensor = torch.FloatTensor([temperature] *
                                            batch_size).to(device)
    top_ps_tensor = torch.FloatTensor([top_p] * batch_size).to(device)
    top_ks_tensor = torch.LongTensor([top_k] * batch_size).to(device)
    output_index = torch.tensor(min_prompt_len, dtype=torch.int64).to(device)

    # Prefill up to min_prompt_len tokens, then treat other prefill as
    # decode and ignore output.
    for i in range(max_seq_len - min_prompt_len):
      next_token_ids = self(
          input_token_ids=input_token_ids_tensor,
          input_positions=input_positions_tensor,
          kv_write_indices=None,
          kv_caches=kv_caches,
          mask=curr_mask_tensor,
          output_positions=output_positions_tensor,
          temperatures=temperatures_tensor,
          top_ps=top_ps_tensor,
          top_ks=top_ks_tensor,
      )

      curr_prompt_mask = prompt_mask_tensor.index_select(
          1, output_index).squeeze(dim=1)
      curr_token_ids = token_ids_tensor.index_select(
          1, output_index).squeeze(dim=1)
      output_token_ids = torch.where(curr_prompt_mask, curr_token_ids,
                                     next_token_ids).unsqueeze(dim=1)
      token_ids_tensor.index_copy_(1, output_index, output_token_ids)

      input_token_ids_tensor = output_token_ids
      input_positions_tensor = output_index.unsqueeze(dim=-1)
      curr_mask_tensor = mask_tensor.index_select(2, input_positions_tensor)
      output_positions_tensor = torch.tensor(0, dtype=torch.int64).to(device)
      output_index = output_index + 1

    # Detokenization.
    token_ids = token_ids_tensor.tolist()
    results = []
    for i, tokens in enumerate(token_ids):
      trimmed_output = tokens[len(prompt_tokens[i]):len(prompt_tokens[i]) +
                              output_len]
      if self.tokenizer.eos_id in trimmed_output:
        eos_index = trimmed_output.index(self.tokenizer.eos_id)
        trimmed_output = trimmed_output[:eos_index]
      results.append(self.tokenizer.decode(trimmed_output))

    # If a string was provided as input, return a string as output.
    return results[0] if is_str_prompt else results

  def load_weights(self, model_path: str):
    self.load_state_dict(
        torch.load(
            model_path,
            mmap=True,
            weights_only=True,
        )['model_state_dict'],
        strict=False,
    )



"""Flax interop."""

import torch
import torchax as tx
import torchax.interop


class FlaxNNModule(torch.nn.Module):

  def __init__(self, env, flax_module, sample_args, sample_kwargs=None):
    super().__init__()
    prng = env.prng_key
    sample_kwargs = sample_kwargs or {}
    parameter_dict = tx.interop.call_jax(flax_module.init, prng, *sample_args,
                                         **sample_kwargs)

    self._params = self._encode_nested_dict(parameter_dict)

    self._flax_module = flax_module

  def _encode_nested_dict(self, nested_dict):
    child_module = torch.nn.Module()
    for k, v in nested_dict.items():
      if isinstance(v, dict):
        child_module.add_module(k, self._encode_nested_dict(v))
      else:
        child_module.register_parameter(k, torch.nn.Parameter(v))
    return child_module

  def _decode_nested_dict(self, child_module):
    result = dict(child_module.named_parameters(recurse=False))
    for k, v in child_module.named_children():
      result[k] = self._decode_nested_dict(v)
    return result

  def forward(self, *args, **kwargs):
    nested_dict_params = self._decode_nested_dict(self._params)
    return tx.interop.call_jax(self._flax_module.apply, nested_dict_params,
                               *args, **kwargs)



import unittest

import torch
from torch.testing._internal.common_utils import TestCase
from torch.testing._internal.common_methods_invocations import op_db
from torch.testing._internal.common_device_type import (
    instantiate_device_type_tests, ops)
from torch.utils import _pytree as pytree
from torchax import tensor
import torchax

skiplist = {
    "_segment_reduce",
    "bincount",  # NOTE: dtype for int input torch gives float. This is weird.
    "byte",
    "cholesky_solve",
    "geqrf",
    "histogram",  # hard op: AssertionError: Tensor-likes are not close!
    "histogramdd",  # TypeError: histogram requires ndarray or scalar arguments, got <class 'list'> at position 1.
    "index_reduce",
    "linalg.ldl_solve",
    "max_pool2d_with_indices_backward",
    "nn.functional.adaptive_max_pool1d",
    "nn.functional.adaptive_max_pool2d",
    "nn.functional.adaptive_max_pool3d",
    "nn.functional.alpha_dropout",
    "nn.functional.ctc_loss",
    "nn.functional.embedding_bag",
    "nn.functional.fractional_max_pool2d",
    "nn.functional.fractional_max_pool3d",
    "nn.functional.multi_head_attention_forward",
    "normal",
    "ormqr",
    "pca_lowrank",
    "searchsorted",
    "special.airy_ai",
    "special.scaled_modified_bessel_k0",
    "special.scaled_modified_bessel_k1",
    "special.spherical_bessel_j0",
    "special.zeta",
    "unfold_copy",
    "unfold",
}

not_support_ops_list = {
    "chalf",  # Skip due to jax not support complex32 with backend: https://github.com/google/jax/issues/14180
    "__rpow__",  # NOTE: cannot fix because torch test case has undefined behavior
    # such as 0 to negative power.
    "to_sparse",  # We are not supporting sparse tensors yet.
    "nn.functional.rrelu",  # pure torch result match torchax test result, only OpInfo mismatch: https://gist.github.com/ManfeiBai/1a449b15f4e946bfcaa3e5ef86da20f4
}

# These inputs are themselves views
# We cannot know how are the views created so cannot replicate the behavior.
variant_test_name_to_skip = {
    "partial_views",
}

random_ops = {
    'empty',
    'empty_like',
    'empty_permuted',
    'empty_strided',
    'bernoulli',
    'geometric',
    'new_empty',
    'new_empty_strided',
    'randint_like',
    'randn',
    'randn_like',
    'rand',
    'rand_like',
    'uniform',
    'multinomial',
    # Dropout is not deterministic https://pytorch.org/docs/stable/generated/torch.nn.functional.feature_alpha_dropout.html
    'nn.functional.feature_alpha_dropout',
    'cauchy',
    'exponential',
    'log_normal',
    'randint',
    'nn.functional.dropout2d',
    'nn.functional.dropout3d',
    'nn.functional.dropout',
}

atol_dict = {
    "cov": (2e-1, 2e-4),
    "linalg.eig": (2e0, 3e0),
    "linalg.eigh": (5e1, 3e0),
    "linalg.eigvalsh": (5e1, 3e0),
    "linalg.pinv": (8e-1, 2e0),
    "linalg.svd": (1e0, 1e0),
    "svd": (1e0, 1e0),
    "svd_lowrank": (1e0, 1e0),
    "matrix_exp": (2e-1, 2e-4),
    "cdist": (5e1, 3e0)
}


def diff_output(testcase,
                output1,
                output2,
                rtol,
                atol,
                equal_nan=True,
                check_output=True):
  if isinstance(output1, torch.Tensor):
    testcase.assertIsInstance(output2, torch.Tensor)
    output2_cpu = output2.detach().cpu()
    if output1.layout != torch.strided:
      # We only compare dense tensors. We dont currently support sparse tensors
      output1 = output1.to_dense()
    if check_output:
      torch.testing.assert_close(
          output2_cpu, output1, rtol=rtol, atol=atol, equal_nan=equal_nan)
    else:
      testcase.assertEqual((output1.shape, output1.dtype),
                           (output2.shape, output2.dtype))
  elif isinstance(output1, (tuple, list)):
    testcase.assertIsInstance(output2, (tuple, list))
    testcase.assertEqual(len(output1), len(output2))
    for o1, o2 in zip(output1, output2):
      diff_output(testcase, o1, o2, rtol, atol)
  else:
    testcase.assertEqual(output1, output2)


def run_export_and_compare(testcase,
                           func,
                           sample_input,
                           check_output=True,
                           equal_nan=True,
                           ignore_indices=False):
  atol, rtol = (1e-3, 1e-5)
  if func.name in atol_dict:
    atol, rtol = atol_dict[func.name]

  with testcase.subTest("torch_eval"):
    res = func(sample_input.input, *sample_input.args, **sample_input.kwargs)
    with testcase.subTest("torchax_eval"):
      input2, args2, kwargs2 = testcase.env.to_xla(
          (sample_input.input, sample_input.args, sample_input.kwargs))
      with testcase.env:
        res2 = func(input2, *args2, **kwargs2)
        res2 = pytree.tree_map_only(tensor.Tensor, lambda t: t.torch(), res2)
        with testcase.subTest("torchax_diff:" + str(atol)):
          if ignore_indices and isinstance(res, tuple) and len(res) == 2:
            diff_output(
                testcase,
                res[0],
                res2[0],
                atol=atol,
                rtol=rtol,
                equal_nan=equal_nan,
                check_output=check_output)
          else:
            diff_output(
                testcase,
                res,
                res2,
                atol=atol,
                rtol=rtol,
                equal_nan=equal_nan,
                check_output=check_output)


ops_to_test = [
    test for test in op_db
    if (test.name not in (skiplist | not_support_ops_list) and
        test.variant_test_name not in variant_test_name_to_skip)
]

# Sort related ops should ignore index;
# For example: sort( [1, 0, 0]) -> [0, 0, 1]
# the correct index can be [1, 2, 0] or [2, 1, 0]
should_ignore_indexes = {"topk", "mode", "kthvalue"}


class TestOpInfo(TestCase):

  @classmethod
  def setUpClass(cls):
    print('op_db size: ', len(op_db), 'testing: ', len(ops_to_test))

  def setUp(self):
    self.env = torchax.default_env()
    torchax.enable_accuracy_mode()
    #self.env.config.debug_accuracy_for_each_op = True
    self.env.config.debug_print_each_op = True
    torch.manual_seed(0)
    self.old_var = self.env.config.use_torch_native_for_cpu_tensor
    self.env.config.use_torch_native_for_cpu_tensor = False

  def tearDown(self):
    self.env.config.use_torch_native_for_cpu_tensor = self.old_var

  # Replaces all values in the input torch_tensor that are less than the given threshold
  # with the threshold value itself.
  def replace_values_below_threshold(self, torch_tensor, threshold):
    return torch.where(torch_tensor < threshold, torch.tensor(threshold),
                       torch_tensor)

  @ops(ops_to_test, allowed_dtypes=(torch.float32, torch.long))
  def test_reference_eager(self, device, dtype, op):
    sample_inputs = op.sample_inputs(device, dtype)
    for sample_input in sample_inputs:
      t = sample_input.input
      if isinstance(t, torch.Tensor) and t.is_sparse:
        continue
      check_output = op.name not in random_ops

      # print("[DEBUG] sample_input: ", sample_input)

      # TODO: this is a workaround to skip int64 cast for linspace
      # reference: https://github.com/pytorch/xla/issues/7505#issuecomment-2400895692 and subsequent comments
      # we have opened a bug in pytorch: https://github.com/pytorch/pytorch/issues/137546
      if op.name == "linspace":
        if 'dtype' in sample_input.kwargs:
          if sample_input.kwargs['dtype'] == torch.int64:
            sample_input.kwargs['dtype'] = torch.float
      if op.name == "polygamma" or op.name == "special.polygamma":
        # The polygamma function is inaccurate for values < 1.
        # To avoid errors during testing, replace values below 1 with 1.
        sample_input.input = self.replace_values_below_threshold(
            sample_input.input, 1)
      if op.name == "nn.functional.scaled_dot_product_attention":
        check_output = sample_input.kwargs.get('dropout_p') == 0.0

      ignore_index = op.name in should_ignore_indexes

      run_export_and_compare(
          self, op, sample_input, check_output, ignore_indices=ignore_index)


instantiate_device_type_tests(TestOpInfo, globals(), only_for={'cpu'})

if __name__ == '__main__':
  unittest.main()



import unittest
import torch
import torchax
from flax import linen as nn
from torchax.flax import FlaxNNModule
from torchax.interop import jax_jit
import jax.numpy as jnp
import jax


class CNN(nn.Module):
  """A simple CNN model."""

  @nn.compact
  def __call__(self, x):
    x = nn.Conv(features=32, kernel_size=(3, 3))(x)
    x = nn.relu(x)
    x = nn.avg_pool(x, window_shape=(2, 2), strides=(2, 2))
    x = nn.Conv(features=64, kernel_size=(3, 3))(x)
    x = nn.relu(x)
    x = nn.avg_pool(x, window_shape=(2, 2), strides=(2, 2))
    x = x.reshape((x.shape[0], -1))  # flatten
    x = nn.Dense(features=256)(x)
    x = nn.relu(x)
    x = nn.Dense(features=10)(x)
    return x


class FlaxTest(unittest.TestCase):

  def test_flax_simple(self):
    flax_model = CNN()

    inputs = jnp.ones((1, 28, 28, 1))
    env = torchax.default_env()
    state = flax_model.init(env.prng_key, inputs)
    expected = flax_model.apply(state, inputs)

    env = torchax.default_env()
    nn_module = FlaxNNModule(env, flax_model, (inputs,), {})
    res = nn_module.forward(inputs)

    self.assertTrue(jnp.allclose(res.jax(), expected))

  def test_flax_functional_call(self):
    flax_model = CNN()

    inputs = jnp.ones((1, 28, 28, 1))
    env = torchax.default_env()
    state = flax_model.init(env.prng_key, inputs)
    expected = flax_model.apply(state, inputs)

    env = torchax.default_env()
    nn_module = FlaxNNModule(env, flax_model, (inputs,), {})

    @jax_jit
    def jitted(weights, args):
      return torch.func.functional_call(nn_module, weights, args)

    with env:
      inputs_torch = torch.ones((1, 28, 28, 1), device='jax')
      state_dict = nn_module.state_dict()
      res = jitted(state_dict, inputs_torch)
      self.assertTrue(jnp.allclose(res.jax(), expected))

  def test_flax_module_nested(self):
    env = torchax.default_env()

    class Parent(torch.nn.Module):

      def __init__(self):
        super().__init__()
        self.a = torch.nn.Linear(28, 28)
        sample_cnn_inputs = torch.ones((1, 28, 28, 1), device='jax')
        self.cnn = FlaxNNModule(env, CNN(), (sample_cnn_inputs,), {})

      def forward(self, x):
        y = self.a(x)
        y = y.reshape((-1, 28, 28, 1))
        res = self.cnn(y)
        return res

    with env:
      nn_module = Parent()

      @jax_jit
      def jitted(weights, args):
        return torch.func.functional_call(nn_module, weights, args)

      inputs_torch = torch.ones((1, 28, 28), device='jax')
      state_dict = nn_module.state_dict()
      res = jitted(state_dict, inputs_torch)
      print(res)


if __name__ == '__main__':
  unittest.main()



import unittest
import torch
import torchax as tx
import torchax.export
import torchax.train
from torch.testing._internal.common_utils import TestCase


class TrainTest(unittest.TestCase):

  def setUp(self):
    torch.manual_seed(0)
    torchax.enable_accuracy_mode()

  def test_scan_module(self):
    x = torch.arange(300).reshape(3, 100).to(torch.float32)
    layers = [
        torch.nn.Linear(100, 100),
        torch.nn.Linear(100, 100),
        torch.nn.Linear(100, 100),
        torch.nn.Linear(100, 100),
    ]
    # repetitively applies the linear
    result = x
    for layer in layers:
      result = layer(result)

    model = tx.train.ScannedModule(layers)

    with torchax.default_env():
      x = x.to('jax')
      model.to('jax')
      result2 = model(x)
      torch.testing.assert_allclose(result, result2.to('cpu'))

  def test_train_step_can_run(self):
    import optax
    with torchax.default_env():
      model = torch.nn.Linear(100, 100)
      model.to('jax')
      weights = model.state_dict()
      x = torch.randn(2, 100).to('jax')
      y = torch.tensor([1, 2]).to('jax')

      def model_fn(weight, buffers, args):
        return torch.func.functional_call(model, weight, args)

      loss_fn = torch.nn.CrossEntropyLoss()

      optimizer = optax.adam(0.01)
      opt_state = tx.interop.call_jax(optimizer.init, weights)

      step = tx.train.make_train_step(model_fn, loss_fn, optimizer)
      print(step(weights, {}, opt_state, x, y))


if __name__ == '__main__':
  unittest.main()



import unittest

import torch
import torchax
from torchax import tensor
import torchax.interop

xla_env = torchax.default_env()


class TestContext(unittest.TestCase):

  def setUp(self):
    self.old_var = xla_env.config.use_torch_native_for_cpu_tensor
    xla_env.config.use_torch_native_for_cpu_tensor = False

  def tearDown(self):
    xla_env.config.use_torch_native_for_cpu_tensor = self.old_var

  def test_mode_context_manager(self):
    with xla_env:
      x = torch.full((3, 3), -1)
      self.assertIsInstance(x, tensor.Tensor)
      y = x.abs()
      self.assertIsInstance(y, tensor.Tensor)

  @staticmethod
  @xla_env
  def _test_mode_decorator():
    x = torch.full((3, 3), -1)
    y = x.abs()

    return x, y

  def test_mode_decorator(self):
    x, y = self._test_mode_decorator()
    self.assertIsInstance(x, tensor.Tensor)
    self.assertIsInstance(y, tensor.Tensor)

  def test_same_manual_seed(self):
    with xla_env:
      xla_env.manual_seed(1234)
      x = torch.randn((3, 3))
      self.assertIsInstance(x, tensor.Tensor)

      xla_env.manual_seed(1234)
      y = torch.randn((3, 3))
      self.assertIsInstance(y, tensor.Tensor)

      self.assertTrue(torch.allclose(x, y))

  def test_different_manual_seed(self):
    with xla_env:
      xla_env.manual_seed(1234)
      x = torch.randn((3, 3))
      self.assertIsInstance(x, tensor.Tensor)

      xla_env.manual_seed(12345)
      y = torch.randn((3, 3))
      self.assertIsInstance(y, tensor.Tensor)

      self.assertFalse(torch.allclose(x, y))

  def test_jit_with_rng(self):

    with xla_env:

      def random_op():
        x = torch.randn(3, 3)
        y = torch.randn(3, 3)
        return x @ y

      random_jit = torchax.interop.jax_jit(random_op)
      self.assertIsInstance(random_jit(), tensor.Tensor)

      # If we run the JIT twice, the random values should be different.
      with self.assertRaises(AssertionError):
        torch.testing.assert_close(random_jit(), random_jit(), atol=0, rtol=0)

  def test_generator_seed(self):
    with xla_env:
      x = torch.randn(2, 3, generator=torch.Generator().manual_seed(0))
      y = torch.randn(2, 3, generator=torch.Generator().manual_seed(0))

      # Values will be the same given the same seed.
      torch.testing.assert_close(x, y)

  def test_buffer(self):

    class M(torch.nn.Module):

      def __init__(self):
        super().__init__()
        c = torch.rand(2)
        self.register_buffer('c', c)
        self.register_buffer('c2', c, persistent=False)

    # Test context manager.
    with xla_env:
      m = M()
      self.assertIsInstance(m.c, tensor.Tensor)
      self.assertIsInstance(m.c2, tensor.Tensor)
    # Test `to_xla`.
    m = M()
    m = xla_env.to_xla(m)
    self.assertIsInstance(m.c, tensor.Tensor)
    self.assertIsInstance(m.c2, tensor.Tensor)


if __name__ == "__main__":
  unittest.main()



from collections.abc import Sequence
from jax._src.numpy.util import promote_dtypes_inexact
import numpy as np
import jax
from jax import numpy as jnp
from jax._src.util import canonicalize_axis
from jax._src import core
from jax._src.image.scale import _kernels, ResizeMethod
from jax import lax
from typing import Callable

# TODO: This block of code needs to be revisited based on https://github.com/jax-ml/jax/issues/24106
# START ----------------- JAX code copied for fixing scale_and_translate -----------------------------

# JAX Link: https://github.com/jax-ml/jax/blob/18f48bd52abe907ff9818da52f3d195d32910c1b/jax/_src/image/scale.py#L52


def compute_weight_mat(input_size: core.DimSize, output_size: core.DimSize,
                       scale, translation, kernel: Callable, antialias: bool):
  dtype = jnp.result_type(scale, translation)
  inv_scale = 1. / scale
  # When downsampling the kernel should be scaled since we want to low pass
  # filter and interpolate, but when upsampling it should not be since we only
  # want to interpolate.
  kernel_scale = jnp.maximum(inv_scale, 1.) if antialias else 1.
  sample_f = ((jnp.arange(output_size, dtype=dtype) + 0.5) * inv_scale -
              translation * inv_scale - 0.5)
  x = (
      jnp.abs(sample_f[jnp.newaxis, :] -
              jnp.arange(input_size, dtype=dtype)[:, jnp.newaxis]) /
      kernel_scale)
  weights = kernel(x)

  total_weight_sum = jnp.sum(weights, axis=0, keepdims=True)
  weights = jnp.where(
      jnp.abs(total_weight_sum) > 1000. * float(np.finfo(np.float32).eps),
      jnp.divide(weights, jnp.where(total_weight_sum != 0, total_weight_sum,
                                    1)), 0)
  # Zero out weights where the sample location is completely outside the input
  # range.
  # Note sample_f has already had the 0.5 removed, hence the weird range below.

  # (barney-s) -------------- returning weights without zeroing ---------------------
  return weights
  input_size_minus_0_5 = core.dimension_as_value(input_size) - 0.5
  return jnp.where(
      jnp.logical_and(sample_f >= -0.5, sample_f
                      <= input_size_minus_0_5)[jnp.newaxis, :], weights, 0)
  # (barney-s) -------------- END returning weights without zeroing ---------------------


# JAX Link: https://github.com/jax-ml/jax/blob/18f48bd52abe907ff9818da52f3d195d32910c1b/jax/_src/image/scale.py#L86


def _scale_and_translate(x, output_shape: core.Shape,
                         spatial_dims: Sequence[int], scale, translation,
                         kernel, antialias: bool, precision):
  input_shape = x.shape
  assert len(input_shape) == len(output_shape)
  assert len(spatial_dims) == len(scale)
  assert len(spatial_dims) == len(translation)
  if len(spatial_dims) == 0:
    return x
  contractions = []
  in_indices = list(range(len(output_shape)))
  out_indices = list(range(len(output_shape)))
  for i, d in enumerate(spatial_dims):
    d = canonicalize_axis(d, x.ndim)
    m = input_shape[d]
    n = output_shape[d]
    w = compute_weight_mat(m, n, scale[i], translation[i], kernel,
                           antialias).astype(x.dtype)
    contractions.append(w)
    contractions.append([d, len(output_shape) + i])
    out_indices[d] = len(output_shape) + i
  contractions.append(out_indices)
  return jnp.einsum(x, in_indices, *contractions, precision=precision)


# JAX Link: https://github.com/jax-ml/jax/blob/18f48bd52abe907ff9818da52f3d195d32910c1b/jax/_src/image/scale.py#L172


# scale and translation here are scalar elements of an np.array, what is the
# correct type annotation?
def scale_and_translate(
    image,
    shape: core.Shape,
    spatial_dims: Sequence[int],
    scale,
    translation,
    # (barney-s) use string
    method: str,  #(barney-s) | ResizeMethod,
    antialias: bool = True,
    precision=lax.Precision.HIGHEST):
  """Apply a scale and translation to an image.

  Generates a new image of shape 'shape' by resampling from the input image
  using the sampling method corresponding to method. For 2D images, this
  operation transforms a location in the input images, (x, y), to a location
  in the output image according to::

    (x * scale[1] + translation[1], y * scale[0] + translation[0])

  (Note the *inverse* warp is used to generate the sample locations.)
  Assumes half-centered pixels, i.e the pixel at integer location ``row, col``
  has coordinates ``y, x = row + 0.5, col + 0.5``, and similarly for other input
  image dimensions.

  If an output location(pixel) maps to an input sample location that is outside
  the input boundaries then the value for the output location will be set to
  zero.

  The ``method`` argument expects one of the following resize methods:

  ``ResizeMethod.LINEAR``, ``"linear"``, ``"bilinear"``, ``"trilinear"``,
    ``"triangle"`` `Linear interpolation`_. If ``antialias`` is ``True``, uses a
    triangular filter when downsampling.

  ``ResizeMethod.CUBIC``, ``"cubic"``, ``"bicubic"``, ``"tricubic"``
    `Cubic interpolation`_, using the Keys cubic kernel.

  ``ResizeMethod.LANCZOS3``, ``"lanczos3"``
    `Lanczos resampling`_, using a kernel of radius 3.

  ``ResizeMethod.LANCZOS5``, ``"lanczos5"``
    `Lanczos resampling`_, using a kernel of radius 5.

  .. _Linear interpolation: https://en.wikipedia.org/wiki/Bilinear_interpolation
  .. _Cubic interpolation: https://en.wikipedia.org/wiki/Bicubic_interpolation
  .. _Lanczos resampling: https://en.wikipedia.org/wiki/Lanczos_resampling

  Args:
    image: a JAX array.
    shape: the output shape, as a sequence of integers with length equal to the
      number of dimensions of `image`.
    spatial_dims: A length K tuple specifying the spatial dimensions that the
      passed scale and translation should be applied to.
    scale: A [K] array with the same number of dimensions as image, containing
      the scale to apply in each dimension.
    translation: A [K] array with the same number of dimensions as image,
      containing the translation to apply in each dimension.
    method: the resizing method to use; either a ``ResizeMethod`` instance or a
      string. Available methods are: LINEAR, LANCZOS3, LANCZOS5, CUBIC.
    antialias: Should an antialiasing filter be used when downsampling? Defaults
      to ``True``. Has no effect when upsampling.

  Returns:
    The scale and translated image.
  """
  shape = core.canonicalize_shape(shape)
  if len(shape) != image.ndim:
    msg = ('shape must have length equal to the number of dimensions of x; '
           f' {shape} vs {image.shape}')
    raise ValueError(msg)
  if isinstance(method, str):
    method = ResizeMethod.from_string(method)
  if method == ResizeMethod.NEAREST:
    # Nearest neighbor is currently special-cased for straight resize, so skip
    # for now.
    raise ValueError('Nearest neighbor resampling is not currently supported '
                     'for scale_and_translate.')
  assert isinstance(method, ResizeMethod)

  kernel = _kernels[method]
  image, = promote_dtypes_inexact(image)
  scale, translation = promote_dtypes_inexact(scale, translation)
  return _scale_and_translate(image, shape, spatial_dims, scale, translation,
                              kernel, antialias, precision)


# END ----------------- END JAX code copied for testing -----------------------------



"""This file contains some decompositons that are not available in torch stable.

Most likely from Content of
https://github.com/pytorch/pytorch/blob/main/torch/_decomp/decompositions.py
at main branch HEAD that we find useful here.

Can also contain decompositions of a torch op in terms of other torch ops.
"""

import functools
from typing import Any, Callable, List, Tuple

import torch
from torch import Tensor
import torch._decomp as decomp
from torch._decomp import decompositions_for_rng
from torch._decomp import register_decomposition
import torch._prims_common as utils
from torch._prims_common.wrappers import out_wrapper

DispatchKey = torch._C.DispatchKey  # type: ignore[attr-defined]

# None of these functions are publicly accessible; get at them
# from torch._decomps
__all__: List[str] = []

aten = torch._ops.ops.aten


def _try_register(op, impl):
  try:
    register_decomposition(op)(impl)
  except:
    pass


@out_wrapper()
def _reflection_pad(a: Tensor, padding: Tuple[int, ...]) -> Tensor:

  def idx(left, middle, right):
    dim_idx = torch.arange(-left, middle + right, device=a.device)
    return middle - 1 - (middle - 1 - dim_idx.abs()).abs()

  return _reflection_or_replication_pad(
      a,
      padding,
      idx,
  )


_try_register(aten.reflection_pad1d, _reflection_pad)
_try_register(aten.reflection_pad2d, _reflection_pad)
_try_register(aten.reflection_pad3d, _reflection_pad)


@out_wrapper()
def _replication_pad(a: Tensor, padding: Tuple[int, ...]) -> Tensor:

  def idx(left, middle, right):
    dim_idx = torch.arange(-left, middle + right, device=a.device)
    return torch.clamp(dim_idx, 0, middle - 1)

  return _reflection_or_replication_pad(
      a,
      padding,
      idx,
  )


decomp.global_decomposition_table["post_autograd"][
    aten.replication_pad2d.default] = _replication_pad


def _reflection_or_replication_pad(
    a: Tensor,
    padding: Tuple[int, ...],
    idx_fn: Callable[[int, int, int], Tensor],
) -> Tensor:
  dim = len(padding) // 2
  torch._check(
      a.dim() in (dim + 1, dim + 2),
      lambda: f"reflection_pad{dim}d requires {dim + 1}D or {dim + 2}D input",
  )
  inp_shape = a.shape[-dim:]
  nc_dim = a.dim() - dim

  padding_left = [padding[2 * (dim - 1 - i)] for i in range(dim)]
  padding_right = [padding[2 * (dim - 1 - i) + 1] for i in range(dim)]

  result = a
  for i in range(dim):
    idx: List[Any] = [None] * result.dim()
    idx[i + nc_dim] = idx_fn(padding_left[i], inp_shape[i], padding_right[i])
    result = aten._unsafe_index(result, idx)

  # convert output to correct memory format, if necessary
  memory_format = utils.suggest_memory_format(result)
  result = result.contiguous(memory_format=memory_format)
  return result


_try_register(aten.replication_pad1d, _replication_pad)
_try_register(aten.replication_pad3d, _replication_pad)


def bernoulli(self, *, generator=None):
  return (torch.rand_like(self, dtype=torch.float32) < self).to(self.dtype)


_try_register(aten.bernoulli.default, bernoulli)


def rand_like(self, **kwargs):
  dtype = kwargs.get("dtype", self.dtype)
  return torch.rand(self.shape, dtype=dtype)


def channel_shuffle(self, groups):
  batchsize, channels, height, width = self.shape
  channels_per_group = channels // groups
  self = self.reshape(batchsize, groups, channels_per_group, height, width)
  self = self.transpose(1, 2)
  self = self.reshape(batchsize, channels, height, width)
  return self


_try_register(aten.channel_shuffle, channel_shuffle)

_try_register(aten.bernoulli, bernoulli)
_try_register(aten.rand_like, rand_like)


def bernoulli_float(self, p=0.5):
  return self.bernoulli_(p)


_try_register(aten.bernoulli_.float, bernoulli_float)
_try_register(aten.bernoulli_.Tensor, decompositions_for_rng.bernoulli_)


def _sum_tensors(ts) -> Tensor:
  return functools.reduce(torch.add, ts)


@register_decomposition(aten.grid_sampler_3d)
def _grid_sampler_3d(
    a: torch.Tensor,
    grid: torch.Tensor,
    interpolation_mode: int = 0,
    padding_mode: int = 0,
    align_corners: bool = False,
) -> Tensor:
  """References: https://github.com/pytorch/pytorch/blob/06a7dc21c1005750598c37f3adbc031183c74de6/torch/_decomp/decompositions.py#L4075

  The above implement the 2d case.
  """
  _expand_grid = False
  torch._check(
      interpolation_mode in (0, 1),
      lambda: f"Invalid interpolation mode {interpolation_mode}",
  )
  torch._check(
      padding_mode in (0, 1, 2), lambda: f"Invalid padding mode {padding_mode}")

  # a is 5D: [B, C, D, H, W]

  def unnormalize(coords: Tensor, size: int) -> Tensor:
    # Rescale coordinates from [-1, 1] to:
    #   [0, size - 1] if align_corners is True
    #   [-.5, size -.5] if align_corners is False
    mul = (size * 0.5 - 0.5) if align_corners else (size * 0.5)
    ofs = size * 0.5 - 0.5
    return coords * mul + ofs

  # Reflects coordinates until they fall between low and high (inclusive).
  # The bounds are passed as twice their value so that half-integer values
  # can be represented as ints.
  def reflect_coordinates(coords: Tensor, twice_low: int,
                          twice_high: int) -> Tensor:
    if twice_low == twice_high:
      return torch.zeros_like(coords)
    coords_min = twice_low / 2
    coords_span = (twice_high - twice_low) / 2
    coords2 = (coords - coords_min).abs()
    extra = torch.fmod(coords2, coords_span)
    flips = (coords2 / coords_span).floor().to(dtype=torch.int8)
    return torch.where(flips & 1 == 0, extra + coords_min,
                       coords_span + coords_min - extra)

  def compute_coordinates(coords: Tensor, size: int) -> Tensor:
    if padding_mode == 0:  # Zero
      return coords
    elif padding_mode == 1:  # Borders
      return torch.clamp(coords, 0, size - 1)
    else:  # padding_mode == 2, Reflection
      if align_corners:
        coords_reflected = reflect_coordinates(coords, 0, 2 * (size - 1))
      else:
        coords_reflected = reflect_coordinates(coords, -1, 2 * size - 1)
      return torch.clamp(coords_reflected, 0, size - 1)

  def compute_source_index(coords: Tensor, size: int) -> Tensor:
    coords_un = unnormalize(coords, size)
    return compute_coordinates(coords_un, size)

  N, C, iD, iH, iW = a.shape
  _, oD, oH, oW, three = grid.shape
  assert three == 3, "Last dim of grid must be 3. got {}".format(three)

  def in_bounds_cond(xs: Tensor, ys: Tensor, zs) -> Tensor:
    xcheck = torch.logical_and(0 <= xs, xs < iW)
    ycheck = torch.logical_and(0 <= ys, ys < iH)
    zcheck = torch.logical_and(0 <= zs, zs < iD)
    return torch.logical_and(xcheck, torch.logical_and(ycheck, zcheck))

  N_idx = torch.arange(N, device=a.device).view(N, 1, 1, 1, 1)
  C_idx = torch.arange(C, device=a.device).view(1, C, 1, 1, 1)

  def clip(xs: torch.Tensor, ys: torch.Tensor, zs, ws: torch.Tensor):
    cond = in_bounds_cond(xs, ys, zs)
    # To clip to inside valid coordinates, we map the coordinates
    # to (x, y) = (0, 0) and also set the weight to 0
    # We also change the shape of the tensor to the appropriate one for
    # broadcasting with N_idx, C_idx for the purposes of advanced indexing
    c = C if _expand_grid else 1
    return tuple(
        torch.where(cond, t, 0).view(N, c, oD, oH, oW) for t in (
            xs.to(dtype=torch.int64),
            ys.to(dtype=torch.int64),
            zs.to(dtype=torch.int64),
            ws,
        ))

  def get_summand(ix: torch.Tensor, iy: torch.Tensor, iz: torch.Tensor,
                  w) -> Tensor:
    # Perform clipping, index into input tensor and multiply by weight
    idx_x, idx_y, idx_z, w_ = clip(ix, iy, iz, w)
    return a[N_idx, C_idx, idx_z, idx_y, idx_x] * w_

  x = grid[..., 0]
  y = grid[..., 1]
  d = grid[..., 2]

  if interpolation_mode == 0:  # Bilinear
    ix = compute_source_index(x, iW)
    iy = compute_source_index(y, iH)
    id_ = compute_source_index(d, iD)

    ix_nwf, iy_nwf, id_nwf = ix.floor(), iy.floor(), id_.floor()
    ix_nef, iy_nef, id_nef = ix_nwf + 1, iy_nwf, id_nwf
    ix_swf, iy_swf, id_swf = ix_nwf, iy_nwf + 1, id_nwf
    ix_sef, iy_sef, id_sef = ix_nef, iy_swf, id_nwf
    ix_nwb, iy_nwb, id_nwb = ix_nwf, iy_nwf, id_nwf + 1
    ix_neb, iy_neb, id_neb = ix_nef, iy_nef, id_nwf + 1
    ix_swb, iy_swb, id_swb = ix_swf, iy_swf, id_nwf + 1
    ix_seb, iy_seb, id_seb = ix_sef, iy_sef, id_nwf + 1

    w_nwf = (ix_seb - ix) * (iy_seb - iy) * (id_seb - id_)
    w_nef = (ix - ix_swb) * (iy_swb - iy) * (id_swb - id_)
    w_swf = (ix_neb - ix) * (iy - iy_neb) * (id_neb - id_)
    w_sef = (ix - ix_nwb) * (iy - iy_nwb) * (id_nwb - id_)
    w_nwb = (ix_sef - ix) * (iy_sef - iy) * (id_ - id_sef)
    w_neb = (ix - ix_swf) * (iy_swf - iy) * (id_ - id_swf)
    w_swb = (ix_nef - ix) * (iy - iy_nef) * (id_ - id_nef)
    w_seb = (ix - ix_nwf) * (iy - iy_nwf) * (id_ - id_nwf)

    return _sum_tensors(
        get_summand(ix, iy, id_, w) for (ix, iy, id_, w) in (
            (ix_nwf, iy_nwf, id_nwf, w_nwf),
            (ix_nef, iy_nef, id_nef, w_nef),
            (ix_swf, iy_swf, id_swf, w_swf),
            (ix_sef, iy_sef, id_sef, w_sef),
            (ix_nwb, iy_nwb, id_nwb, w_nwb),
            (ix_neb, iy_neb, id_neb, w_neb),
            (ix_swb, iy_swb, id_swb, w_swb),
            (ix_seb, iy_seb, id_seb, w_seb),
        ))
  else:  # interpolation_mode == 1:  # Nearest
    ix = compute_source_index(x, iW)
    iy = compute_source_index(y, iH)
    iz = compute_source_index(d, iD)

    ix_nearest = ix.round()
    iy_nearest = iy.round()
    iz_nearest = iz.round()

    return get_summand(ix_nearest, iy_nearest, iz_nearest, 1)


DECOMPOSITIONS = decomp.get_decompositions([
    torch.ops.aten.upsample_bicubic2d,
    torch.ops.aten.upsample_nearest1d,
    torch.ops.aten.upsample_nearest2d,
    torch.ops.aten.upsample_nearest3d,
    torch.ops.aten._upsample_nearest_exact1d,
    torch.ops.aten._upsample_nearest_exact2d,
    torch.ops.aten._upsample_nearest_exact3d,
    torch.ops.aten._native_batch_norm_legit.no_stats,
    torch.ops.aten._native_batch_norm_legit_functional.default,
    torch.ops.aten._adaptive_avg_pool2d,
    torch.ops.aten._adaptive_avg_pool3d,
    torch.ops.aten.grid_sampler_2d,
    torch.ops.aten.grid_sampler_3d,
    torch.ops.aten.native_dropout,
    torch.ops.aten.reflection_pad1d,
    torch.ops.aten.reflection_pad2d,
    torch.ops.aten.reflection_pad3d,
    torch.ops.aten.replication_pad1d,
    torch.ops.aten.replication_pad2d,
    torch.ops.aten.replication_pad3d,
    torch.ops.aten.bernoulli,
    torch.ops.aten.rand_like,
    torch.ops.aten._batch_norm_with_update,
    torch.ops.aten.channel_shuffle,
    torch.ops.aten.nll_loss2d_forward,
    torch.ops.aten.nll_loss2d_backward,
    torch.ops.aten.bernoulli_.Tensor,
    torch.ops.aten.bernoulli_.float,
    torch.ops.aten.log_normal,
    torch.ops.aten.addcdiv.default,
    torch.ops.aten.addcdiv.out,
    torch.ops.aten.addcdiv_.default,
    torch.ops.aten.addcmul.default,
    torch.ops.aten.addcmul.out,
    torch.ops.aten.addcmul_.default,
    torch.ops.aten.addr.default,
    torch.ops.aten.addr.out,
    torch.ops.aten.affine_grid_generator.default,
    torch.ops.aten.affine_grid_generator.out,
    torch.ops.aten.alias_copy.default,
    torch.ops.aten.alias_copy.out,
    torch.ops.aten.all.default,
    torch.ops.aten.all.dim,
    torch.ops.aten.all.dims,
    torch.ops.aten.all.out,
    torch.ops.aten.all.dims_out,
    torch.ops.aten.all.all_out,
    torch.ops.aten.all.dimname,
    torch.ops.aten.all.dimname_out,
    torch.ops.aten.aminmax.default,
    torch.ops.aten.aminmax.out,
    torch.ops.aten.arange.default,
    torch.ops.aten.arange.start,
    torch.ops.aten.baddbmm.default,
    torch.ops.aten.baddbmm.out,
    torch.ops.aten.binary_cross_entropy.default,
    torch.ops.aten.binary_cross_entropy.out,
    torch.ops.aten.binary_cross_entropy_backward.default,
    torch.ops.aten.binary_cross_entropy_backward.grad_input,
    torch.ops.aten.binary_cross_entropy_with_logits.default,
    torch.ops.aten.binary_cross_entropy_with_logits.out,
    torch.ops.aten.block_diag.default,
    torch.ops.aten.block_diag.out,
    torch.ops.aten.celu.default,
    torch.ops.aten.celu.out,
    torch.ops.aten.celu_.default,
    torch.ops.aten.channel_shuffle.default,
    torch.ops.aten.channel_shuffle.out,
    torch.ops.aten.clamp_max.default,
    torch.ops.aten.clamp_max.Tensor,
    torch.ops.aten.clamp_max.out,
    torch.ops.aten.clamp_max.Tensor_out,
    torch.ops.aten.clamp_min.default,
    torch.ops.aten.clamp_min.Tensor,
    torch.ops.aten.clamp_min.out,
    torch.ops.aten.clamp_min.Tensor_out,
    torch.ops.aten.col2im.default,
    torch.ops.aten.col2im.out,
    torch.ops.aten.count_nonzero.dim_IntList,
    torch.ops.aten.count_nonzero.dim_IntList_out,
    torch.ops.aten.count_nonzero.default,
    torch.ops.aten.count_nonzero.out,
    torch.ops.aten.linalg_cross.default,
    torch.ops.aten.linalg_cross.out,
    torch.ops.aten.cudnn_batch_norm.default,
    torch.ops.aten.cudnn_batch_norm.out,
    torch.ops.aten.cudnn_batch_norm_backward.default,
    torch.ops.aten.cudnn_batch_norm_backward.out,
    torch.ops.aten.miopen_batch_norm_backward.default,
    torch.ops.aten.miopen_batch_norm_backward.out,
    torch.ops.aten.deg2rad.default,
    torch.ops.aten.deg2rad.out,
    torch.ops.aten.deg2rad_.default,
    torch.ops.aten.detach.default,
    torch.ops.aten.diag_embed.default,
    torch.ops.aten.diag_embed.out,
    torch.ops.aten.diagonal_backward.default,
    torch.ops.aten.diagonal_backward.out,
    torch.ops.aten.dot.default,
    torch.ops.aten.dot.out,
    torch.ops.aten.vdot.default,
    torch.ops.aten.vdot.out,
    torch.ops.aten.elu.default,
    torch.ops.aten.elu.out,
    torch.ops.aten.elu_.default,
    torch.ops.aten.elu_backward.default,
    torch.ops.aten.elu_backward.grad_input,
    torch.ops.aten.embedding_dense_backward.default,
    torch.ops.aten.embedding_dense_backward.out,
    torch.ops.aten.empty_like.default,
    torch.ops.aten.empty_like.out,
    torch.ops.aten._euclidean_dist.default,
    torch.ops.aten.expand_copy.default,
    torch.ops.aten.expand_copy.out,
    torch.ops.aten.eye.default,
    torch.ops.aten.eye.m,
    torch.ops.aten.eye.out,
    torch.ops.aten.eye.m_out,
    torch.ops.aten.fill.Scalar,
    torch.ops.aten.fill.Tensor,
    torch.ops.aten.fill_.Scalar,
    torch.ops.aten.fill_.Tensor,
    torch.ops.aten.floor_divide.default,
    torch.ops.aten.floor_divide.Scalar,
    torch.ops.aten.floor_divide.out,
    torch.ops.aten.floor_divide.Scalar_out,
    torch.ops.aten.frac.default,
    torch.ops.aten.frac.out,
    torch.ops.aten.frac_.default,
    torch.ops.aten.gelu_.default,
    torch.ops.aten.gelu_backward.default,
    torch.ops.aten.gelu_backward.grad_input,
    torch.ops.aten.glu.default,
    torch.ops.aten.glu.out,
    torch.ops.aten.glu_backward.default,
    torch.ops.aten.glu_backward.grad_input,
    torch.ops.aten.hardshrink.default,
    torch.ops.aten.hardshrink.out,
    torch.ops.aten.hardsigmoid.default,
    torch.ops.aten.hardsigmoid.out,
    torch.ops.aten.hardsigmoid_.default,
    torch.ops.aten.hardsigmoid_backward.default,
    torch.ops.aten.hardsigmoid_backward.grad_input,
    torch.ops.aten.hardswish.default,
    torch.ops.aten.hardswish.out,
    torch.ops.aten.hardswish_.default,
    torch.ops.aten.hardswish_backward.default,
    torch.ops.aten.hardswish_backward.out,
    torch.ops.aten.hardtanh_.default,
    torch.ops.aten.hardtanh_backward.default,
    torch.ops.aten.hardtanh_backward.grad_input,
    torch.ops.aten.heaviside.default,
    torch.ops.aten.heaviside.out,
    torch.ops.aten.heaviside_.default,
    torch.ops.aten.huber_loss.default,
    torch.ops.aten.huber_loss.out,
    torch.ops.aten.huber_loss_backward.default,
    torch.ops.aten.huber_loss_backward.out,
    torch.ops.aten.im2col.default,
    torch.ops.aten.im2col.out,
    torch.ops.aten.index_add.default,
    torch.ops.aten.index_add.out,
    torch.ops.aten.index_add.dimname,
    torch.ops.aten.index_add_.default,
    torch.ops.aten.index_copy.default,
    torch.ops.aten.index_copy.dimname,
    torch.ops.aten.index_copy.out,
    torch.ops.aten.index_copy_.default,
    torch.ops.aten.index_copy_.dimname,
    torch.ops.aten.index_fill.int_Tensor,
    torch.ops.aten.index_fill.int_Scalar,
    torch.ops.aten.index_fill.Dimname_Scalar,
    torch.ops.aten.index_fill.Dimname_Tensor,
    torch.ops.aten.index_fill.int_Scalar_out,
    torch.ops.aten.index_fill.int_Tensor_out,
    torch.ops.aten.index_fill_.int_Tensor,
    torch.ops.aten.index_fill_.int_Scalar,
    torch.ops.aten.index_fill_.Dimname_Scalar,
    torch.ops.aten.index_fill_.Dimname_Tensor,
    torch.ops.aten.isin.Tensor_Tensor,
    torch.ops.aten.isin.Tensor_Tensor_out,
    torch.ops.aten.isin.Tensor_Scalar,
    torch.ops.aten.isin.Tensor_Scalar_out,
    torch.ops.aten.isin.Scalar_Tensor,
    torch.ops.aten.isin.Scalar_Tensor_out,
    torch.ops.aten.isneginf.default,
    torch.ops.aten.isneginf.out,
    torch.ops.aten.isposinf.default,
    torch.ops.aten.isposinf.out,
    torch.ops.aten.leaky_relu_.default,
    torch.ops.aten.leaky_relu_backward.default,
    torch.ops.aten.leaky_relu_backward.grad_input,
    torch.ops.aten.lerp.Scalar,
    torch.ops.aten.lerp.Tensor,
    torch.ops.aten.lerp.Scalar_out,
    torch.ops.aten.lerp.Tensor_out,
    torch.ops.aten.lerp_.Scalar,
    torch.ops.aten.lerp_.Tensor,
    torch.ops.aten.linspace.Tensor_Tensor,
    torch.ops.aten.linspace.Tensor_Scalar,
    torch.ops.aten.linspace.Scalar_Tensor,
    torch.ops.aten.linspace.default,
    torch.ops.aten.linspace.out,
    torch.ops.aten.linspace.Tensor_Tensor_out,
    torch.ops.aten.linspace.Tensor_Scalar_out,
    torch.ops.aten.linspace.Scalar_Tensor_out,
    torch.ops.aten.logaddexp.default,
    torch.ops.aten.logaddexp.out,
    torch.ops.aten.logaddexp2.default,
    torch.ops.aten.logaddexp2.out,
    torch.ops.aten.logit.default,
    torch.ops.aten.logit.out,
    torch.ops.aten.logit_.default,
    torch.ops.aten.logit_backward.default,
    torch.ops.aten.log_sigmoid_backward.default,
    torch.ops.aten.log_sigmoid_backward.grad_input,
    torch.ops.aten.log_sigmoid_forward.default,
    torch.ops.aten.log_sigmoid_forward.output,
    torch.ops.aten._log_softmax_backward_data.default,
    torch.ops.aten._log_softmax_backward_data.out,
    torch.ops.aten.logspace.Tensor_Tensor,
    torch.ops.aten.logspace.Tensor_Scalar,
    torch.ops.aten.logspace.Scalar_Tensor,
    torch.ops.aten.logspace.default,
    torch.ops.aten.logspace.out,
    torch.ops.aten.logspace.Tensor_Tensor_out,
    torch.ops.aten.logspace.Tensor_Scalar_out,
    torch.ops.aten.logspace.Scalar_Tensor_out,
    torch.ops.aten.logsumexp.default,
    torch.ops.aten.masked_fill.Scalar,
    torch.ops.aten.masked_fill.Tensor,
    torch.ops.aten.masked_fill.Scalar_out,
    torch.ops.aten.masked_fill.Tensor_out,
    torch.ops.aten.masked_fill_.Scalar,
    torch.ops.aten.masked_fill_.Tensor,
    torch.ops.aten.mish.default,
    torch.ops.aten.mish.out,
    torch.ops.aten.mish_.default,
    torch.ops.aten.mse_loss.default,
    torch.ops.aten.mse_loss.out,
    torch.ops.aten.mse_loss_backward.default,
    torch.ops.aten.mse_loss_backward.grad_input,
    torch.ops.aten.multi_margin_loss.default,
    torch.ops.aten.multi_margin_loss.out,
    torch.ops.aten.multilabel_margin_loss_forward.default,
    torch.ops.aten.multilabel_margin_loss_forward.output,
    torch.ops.aten.mv.default,
    torch.ops.aten.mv.out,
    torch.ops.aten.mvlgamma.default,
    torch.ops.aten.mvlgamma.out,
    torch.ops.aten.mvlgamma_.default,
    torch.ops.aten.nansum.default,
    torch.ops.aten.nansum.out,
    torch.ops.aten.nan_to_num.default,
    torch.ops.aten.nan_to_num.out,
    torch.ops.aten.nan_to_num_.default,
    torch.ops.aten.native_batch_norm_backward.default,
    torch.ops.aten.native_batch_norm_backward.out,
    torch.ops.aten.native_dropout_backward.default,
    torch.ops.aten.native_dropout_backward.out,
    torch.ops.aten.native_group_norm_backward.default,
    torch.ops.aten.native_group_norm_backward.out,
    torch.ops.aten.native_layer_norm_backward.default,
    torch.ops.aten.native_layer_norm_backward.out,
    torch.ops.aten.new_empty.default,
    torch.ops.aten.new_empty.out,
    torch.ops.aten.new_full.default,
    torch.ops.aten.new_full.out,
    torch.ops.aten.new_ones.default,
    torch.ops.aten.new_ones.out,
    torch.ops.aten.new_zeros.default,
    torch.ops.aten.new_zeros.out,
    torch.ops.aten.nll_loss2d_forward.default,
    torch.ops.aten.nll_loss2d_forward.output,
    torch.ops.aten.nll_loss2d_backward.default,
    torch.ops.aten.nll_loss2d_backward.grad_input,
    torch.ops.aten.nll_loss_backward.default,
    torch.ops.aten.nll_loss_backward.grad_input,
    torch.ops.aten.nll_loss_forward.default,
    torch.ops.aten.nll_loss_forward.output,
    torch.ops.aten.norm.Scalar,
    torch.ops.aten.norm.ScalarOpt_dim,
    torch.ops.aten.norm.names_ScalarOpt_dim,
    torch.ops.aten.norm.ScalarOpt_dim_dtype,
    torch.ops.aten.norm.dtype_out,
    torch.ops.aten.norm.out,
    torch.ops.aten.norm.ScalarOpt_dtype,
    torch.ops.aten.norm.ScalarOpt_dtype_out,
    torch.ops.aten.norm.Scalar_out,
    torch.ops.aten.norm.names_ScalarOpt_dim_dtype,
    torch.ops.aten.norm.names_dtype_out,
    torch.ops.aten.norm.names_out,
    torch.ops.aten.ones.default,
    torch.ops.aten.ones_like.default,
    torch.ops.aten.ones_like.out,
    torch.ops.aten.pixel_shuffle.default,
    torch.ops.aten.pixel_shuffle.out,
    torch.ops.aten.pixel_unshuffle.default,
    torch.ops.aten.pixel_unshuffle.out,
    torch.ops.aten._prelu_kernel.default,
    torch.ops.aten._prelu_kernel_backward.default,
    torch.ops.aten._reshape_alias.default,
    torch.ops.aten.rad2deg.default,
    torch.ops.aten.rad2deg.out,
    torch.ops.aten.rad2deg_.default,
    torch.ops.aten.reflection_pad1d.default,
    torch.ops.aten.reflection_pad1d.out,
    torch.ops.aten.reflection_pad1d_backward.default,
    torch.ops.aten.reflection_pad1d_backward.grad_input,
    torch.ops.aten.reflection_pad2d.default,
    torch.ops.aten.reflection_pad2d.out,
    torch.ops.aten.reflection_pad2d_backward.default,
    torch.ops.aten.reflection_pad2d_backward.grad_input,
    torch.ops.aten.reflection_pad3d.default,
    torch.ops.aten.reflection_pad3d.out,
    torch.ops.aten.reflection_pad3d_backward.default,
    torch.ops.aten.reflection_pad3d_backward.grad_input,
    torch.ops.aten.replication_pad1d.default,
    torch.ops.aten.replication_pad1d.out,
    torch.ops.aten.replication_pad2d.default,
    torch.ops.aten.replication_pad2d.out,
    torch.ops.aten.replication_pad3d.default,
    torch.ops.aten.replication_pad3d.out,
    torch.ops.aten.renorm.default,
    torch.ops.aten.renorm.out,
    torch.ops.aten.renorm_.default,
    torch.ops.aten.resize_as.default,
    torch.ops.aten.resize_as.out,
    torch.ops.aten.roll.default,
    torch.ops.aten.roll.out,
    torch.ops.aten.rot90.default,
    torch.ops.aten.rot90.out,
    torch.ops.aten.rrelu_with_noise.default,
    torch.ops.aten.rrelu_with_noise.out,
    torch.ops.aten.rrelu_with_noise_.default,
    torch.ops.aten.rsub.Tensor,
    torch.ops.aten.rsub.Scalar,
    torch.ops.aten.rsub.Tensor_out,
    torch.ops.aten.rsub.Scalar_out,
    torch.ops.aten._safe_softmax.default,
    torch.ops.aten._scaled_dot_product_flash_attention_for_cpu.default,
    torch.ops.aten.select_backward.default,
    torch.ops.aten.select_backward.out,
    torch.ops.aten.select_scatter.default,
    torch.ops.aten.select_scatter.out,
    torch.ops.aten.sgn.default,
    torch.ops.aten.sgn.out,
    torch.ops.aten.sgn_.default,
    torch.ops.aten.sigmoid_backward.default,
    torch.ops.aten.sigmoid_backward.grad_input,
    torch.ops.aten.silu.default,
    torch.ops.aten.silu.out,
    torch.ops.aten.silu_.default,
    torch.ops.aten.silu_backward.default,
    torch.ops.aten.silu_backward.grad_input,
    torch.ops.aten.sinc.default,
    torch.ops.aten.sinc.out,
    torch.ops.aten.sinc_.default,
    torch.ops.aten.slice_backward.default,
    torch.ops.aten.slice_backward.out,
    torch.ops.aten.smooth_l1_loss.default,
    torch.ops.aten.smooth_l1_loss.out,
    torch.ops.aten.smooth_l1_loss_backward.default,
    torch.ops.aten.smooth_l1_loss_backward.grad_input,
    torch.ops.aten.soft_margin_loss.default,
    torch.ops.aten.soft_margin_loss.out,
    torch.ops.aten.soft_margin_loss_backward.default,
    torch.ops.aten.soft_margin_loss_backward.grad_input,
    torch.ops.aten._softmax_backward_data.default,
    torch.ops.aten._softmax_backward_data.out,
    torch.ops.aten.softplus.default,
    torch.ops.aten.softplus.out,
    torch.ops.aten.softplus_backward.default,
    torch.ops.aten.softplus_backward.grad_input,
    torch.ops.aten.softshrink.default,
    torch.ops.aten.softshrink.out,
    torch.ops.aten.special_entr.default,
    torch.ops.aten.special_entr.out,
    torch.ops.aten.special_log_ndtr.default,
    torch.ops.aten.special_log_ndtr.out,
    torch.ops.aten.special_xlog1py.default,
    torch.ops.aten.special_xlog1py.other_scalar,
    torch.ops.aten.special_xlog1py.self_scalar,
    torch.ops.aten.special_xlog1py.out,
    torch.ops.aten.special_xlog1py.self_scalar_out,
    torch.ops.aten.special_xlog1py.other_scalar_out,
    torch.ops.aten.split.Tensor,
    torch.ops.aten.split_with_sizes_copy.default,
    torch.ops.aten.split_with_sizes_copy.out,
    torch.ops.aten.squeeze.default,
    torch.ops.aten.squeeze.dim,
    torch.ops.aten.std.default,
    torch.ops.aten.std.dim,
    torch.ops.aten.std.correction,
    torch.ops.aten.std.names_dim,
    torch.ops.aten.std.names_out,
    torch.ops.aten.std.out,
    torch.ops.aten.std.correction_out,
    torch.ops.aten.std.correction_names,
    torch.ops.aten.std.correction_names_out,
    torch.ops.aten.std_mean.default,
    torch.ops.aten.std_mean.dim,
    torch.ops.aten.std_mean.correction,
    torch.ops.aten.std_mean.names_dim,
    torch.ops.aten.std_mean.correction_names,
    torch.ops.aten.std_mean.correction_out,
    torch.ops.aten.stack.default,
    torch.ops.aten.stack.out,
    torch.ops.aten.sum.default,
    torch.ops.aten.sum.out,
    torch.ops.aten.t.default,
    torch.ops.aten.t_copy.out,
    torch.ops.aten.t_copy.default,
    torch.ops.aten.take.default,
    torch.ops.aten.take.out,
    torch.ops.aten.tanh_backward.default,
    torch.ops.aten.tanh_backward.grad_input,
    torch.ops.aten.threshold.default,
    torch.ops.aten.threshold.out,
    torch.ops.aten.threshold_.default,
    torch.ops.aten.threshold_backward.default,
    torch.ops.aten.threshold_backward.grad_input,
    torch.ops.aten.trace.default,
    torch.ops.aten.trace.out,
    torch.ops.aten.transpose.int,
    torch.ops.aten.tril.default,
    torch.ops.aten.tril.out,
    torch.ops.aten.tril_.default,
    torch.ops.aten.triu.default,
    torch.ops.aten.triu.out,
    torch.ops.aten.triu_.default,
    torch.ops.aten.unbind.int,
    torch.ops.aten.unbind.Dimname,
    torch.ops.aten.unfold_backward.default,
    torch.ops.aten.unfold_backward.out,
    torch.ops.aten.unfold_copy.default,
    torch.ops.aten.unfold_copy.out,
    torch.ops.aten._unsafe_index.Tensor,
    torch.ops.aten._unsafe_index_put.default,
    torch.ops.aten._unsafe_masked_index.default,
    torch.ops.aten._unsafe_masked_index_put_accumulate.default,
    torch.ops.aten.unsafe_split.Tensor,
    torch.ops.aten.unsafe_split_with_sizes.default,
    torch.ops.aten.unsqueeze_copy.out,
    torch.ops.aten.unsqueeze_copy.default,
    torch.ops.aten._unsafe_view.default,
    torch.ops.aten._unsafe_view.out,
    torch.ops.aten.upsample_linear1d.default,
    torch.ops.aten.upsample_linear1d.out,
    torch.ops.aten.upsample_bilinear2d.vec,
    torch.ops.aten.upsample_bilinear2d.default,
    torch.ops.aten.upsample_bilinear2d.out,
    torch.ops.aten.upsample_trilinear3d.vec,
    torch.ops.aten.upsample_trilinear3d.default,
    torch.ops.aten.upsample_trilinear3d.out,
    torch.ops.aten.xlogy.Tensor,
    torch.ops.aten.xlogy.Scalar_Other,
    torch.ops.aten.xlogy.Scalar_Self,
    torch.ops.aten.xlogy.OutTensor,
    torch.ops.aten.xlogy.OutScalar_Self,
    torch.ops.aten.xlogy.OutScalar_Other,
    torch.ops.aten.xlogy_.Tensor,
    torch.ops.aten.xlogy_.Scalar_Other,
    torch.ops.aten.zero.default,
    torch.ops.aten.zero.out,
    torch.ops.aten.zero_.default,
    torch.ops.aten.zeros.default,
    torch.ops.aten.zeros_like.default,
    torch.ops.aten.zeros_like.out,
    torch.ops.aten._chunk_cat.default,
    torch.ops.aten._chunk_cat.out,
    torch.ops.aten._weight_norm_interface.default,
    torch.ops.aten._weight_norm_interface.out,
    torch.ops.aten.__iand__.Tensor,
    torch.ops.aten.__ixor__.Tensor,
    torch.ops.aten.__ilshift__.Tensor,
    torch.ops.aten.__ilshift__.Scalar,
    torch.ops.aten.__irshift__.Tensor,
    torch.ops.aten.__irshift__.Scalar,
    torch.ops.aten.__ior__.Tensor,
])

MUTABLE_DECOMPOSITION = [
    torch.ops.aten.bernoulli_.Tensor,
    torch.ops.aten.bernoulli_.float,
]



## Please read: https://jax.readthedocs.io/en/latest/notebooks/thinking_in_jax.html

## Goal 1. Jax jit what it is
## Goal 2. Illustrate that changing shapes == recompile
## Goal 3. Illustrate that closures == inlined constant in graph (undesirable)

import random
import time
import jax.numpy as jnp
import numpy as np
from jax import jit
import jax


def norm(X):
  for i in range(10):
    X = X @ X
    X = X - X.mean(0)
  return X / X.std(0)


norm_compiled = jit(norm)

np.random.seed(1701)

# print('---- example 1 -----')
# for i in range(5):
#   X = jnp.array(np.random.rand(1000, 1000))
#   start = time.perf_counter()
#   norm(X).block_until_ready()
#   end = time.perf_counter()
#   print(f'iteration {i}: norm: {end - start}')
#   start = time.perf_counter()
#   norm_compiled(X).block_until_ready()
#   end = time.perf_counter()
#   print(f'iteration {i}: norm_compiled: {end - start}')

#jax.config.update("jax_explain_cache_misses", True)
#print('---- example 3 -----')
#print('---- example 2 -----')
# for i in range(5):
#   shape = random.randint(1000, 2000)
#   print('shape is ', shape)
#   X = jnp.array(np.random.rand(shape, shape))
#   start = time.perf_counter()
#   norm(X).block_until_ready()
#   end = time.perf_counter()
#   print(f'iteration {i}: norm: {end - start}')
#   start = time.perf_counter()
#   norm_compiled(X).block_until_ready()
#   end = time.perf_counter()
#   print(f'iteration {i}: norm_compiled: {end - start}')

#Example 4: print out the graph
print('--- example 4 ---')

X = jnp.array(np.random.rand(1000, 1000))

#print(norm_compiled.lower(X).as_text())
#print(norm_compiled.lower(jax.ShapeDtypeStruct((1000, 1000), jnp.float32.dtype)).as_text())

# Example 5: What happen to closures
print('--- example 5 ---')


def addx(y):
  #print('y is ', y)
  #print('X is ', X)
  # import pdb; pdb.set_trace()
  #jax.debug.print(...)
  #jax.debug.breakpoint()
  return y + X


addx_jitted = jax.jit(addx).lower(
    jax.ShapeDtypeStruct((1000, 1000), jnp.float32.dtype))
#print(addx_jitted.as_text())
#print(addx_jitted.compile()(X))

# ASIDE: pdb; print;
# https://jax.readthedocs.io/en/latest/debugging/print_breakpoint.html

# Example 6: What happens with class attr


class Model:

  def __init__(self):
    self.weight = jnp.array(np.random.randn(1000, 1000))
    self.bias = jnp.array(np.random.randn(1000))

  def __call__(self, X):
    return X @ self.weight + self.bias


m = Model()
print('Not jitted', m(X))
print(jax.jit(m).lower(X).as_text())


def model_pure(weight, bias, X):
  m.weight = weight
  m.bias = bias
  res = m(X)
  return res



import torch
from torch import nn
from torch.nn import functional as F
from torch.utils import _pytree as pytree


class Subclass(torch.Tensor):

  def __new__(cls, raw_data, requires_grad=False):
    return torch.Tensor._make_subclass(
        cls,
        raw_data,
        require_grad=requires_grad,
    )

  def __init__(self, raw_data=None, requires_grad=False):
    # Store any provided user raw_data
    self.raw_data = raw_data

  @classmethod
  #def __torch_dispatch__(cls, func, types, args=(), kwargs=None):
  def __torch_function__(cls, func, types, args=(), kwargs=None):
    kwargs = kwargs or {}
    print(f'func is {func}')

    def unpack(x):
      if isinstance(x, Subclass):
        return x.raw_data
      return x

    (args, kwargs) = pytree.tree_map(unpack, (args, kwargs))
    res = func(*args, **kwargs)
    return pytree.tree_map_only(torch.Tensor, Subclass, res)

  def __str__(self):
    return f'Subclass of shape {self.shape}'

  def add(self, a):
    print('HERE: add')
    return super().add(a)

  __repr__ = __str__


class MyModel(nn.Module):

  def __init__(self):
    super().__init__()
    self.fc1 = nn.Linear(28 * 28, 120)
    self.fc2 = nn.Linear(120, 84)
    self.fc3 = nn.Linear(84, 10)

  def forward(self, x):
    x = x.view(-1, 28 * 28)
    x = F.relu(self.fc1(x))
    x = F.relu(self.fc2(x))
    x = self.fc3(x)
    return x


model = MyModel()

x = torch.randn(10, 28 * 28)
x2 = Subclass(x)
print(model(x2))

x2.add(2)
